@@ -1,5 +1,6 @@
 from __future__ import with_statement
 from collections import defaultdict, namedtuple
+from functools import partial
 from operator import methodcaller
 import os
 import re
@@ -8,9 +9,10 @@ import copy
 
 from pytest import raises
 
-from schema import (Schema, Use, And, Or, Regex, Optional,
+from schema import (Schema, Use, And, Or, Regex, Optional, Const,
                     SchemaError, SchemaWrongKeyError,
-                    SchemaMissingKeyError, SchemaUnexpectedTypeError)
+                    SchemaMissingKeyError, SchemaUnexpectedTypeError,
+                    SchemaForbiddenKeyError, Forbidden)
 
 if sys.version_info[0] == 3:
     basestring = str  # Python 3 does not have basestring
@@ -75,6 +77,28 @@ def test_or():
     with SE: Or().validate(2)
 
 
+def test_test():
+    def unique_list(_list):
+        return len(_list) == len(set(_list))
+
+    def dict_keys(key, _list):
+        return list(map(lambda d: d[key], _list))
+
+    schema = (
+        Schema(
+            Const(
+                And(Use(partial(dict_keys, "index")), unique_list))))
+    data = [
+        {"index": 1, "value": "foo"},
+        {"index": 2, "value": "bar"}]
+    assert schema.validate(data) == data
+
+    bad_data = [
+        {"index": 1, "value": "foo"},
+        {"index": 1, "value": "bar"}]
+    with SE: schema.validate(bad_data)
+
+
 def test_regex():
     # Simple case: validate string
     assert Regex(r'foo').validate('afoot') == 'afoot'
@@ -200,11 +224,34 @@ def test_dict_keys():
 def test_ignore_extra_keys():
     assert Schema({'key': 5}, ignore_extra_keys=True).validate(
             {'key': 5, 'bad': 4}) == {'key': 5}
+    assert Schema({'key': 5, 'dk': {'a': 'a'}}, ignore_extra_keys=True).validate(
+        {'key': 5, 'bad': 'b', 'dk': {'a': 'a', 'bad': 'b'}}) == \
+        {'key': 5, 'dk': {'a': 'a'}}
+    assert Schema([{'key': 'v'}], ignore_extra_keys=True).validate(
+        [{'key': 'v', 'bad': 'bad'}]) == [{'key': 'v'}]
+    assert Schema([{'key': 'v'}], ignore_extra_keys=True).validate(
+        [{'key': 'v', 'bad': 'bad'}]) == [{'key': 'v'}]
 
 
 def test_ignore_extra_keys_validation_and_return_keys():
     assert Schema({'key': 5, object: object}, ignore_extra_keys=True).validate(
             {'key': 5, 'bad': 4}) == {'key': 5, 'bad': 4}
+    assert Schema({'key': 5, 'dk': {'a': 'a', object: object}},
+                  ignore_extra_keys=True).validate(
+        {'key': 5, 'dk': {'a': 'a', 'bad': 'b'}}) == \
+        {'key': 5, 'dk': {'a': 'a', 'bad': 'b'}}
+
+
+def test_dict_forbidden_keys():
+    with raises(SchemaForbiddenKeyError):
+        Schema({Forbidden('b'): object}).validate({'b': 'bye'})
+    with raises(SchemaWrongKeyError):
+        Schema({Forbidden('b'): int}).validate({'b': 'bye'})
+    assert (Schema({Forbidden('b'): int,
+                   Optional('b'): object}).validate({'b': 'bye'}) ==
+            {'b': 'bye'})
+    with raises(SchemaForbiddenKeyError):
+        Schema({Forbidden('b'): object, Optional('b'): object}).validate({'b': 'bye'})
 
 
 def test_dict_optional_keys():
@@ -215,6 +262,8 @@ def test_dict_optional_keys():
     # Make sure Optionals are favored over types:
     assert Schema({basestring: 1,
                    Optional('b'): 2}).validate({'a': 1, 'b': 2}) == {'a': 1, 'b': 2}
+    # Make sure Optionals hash based on their key:
+    assert len({Optional('a'): 1, Optional('a'): 1, Optional('b'): 2}) == 2
 
 
 def test_dict_optional_defaults():
@@ -240,6 +289,22 @@ def test_dict_subtypes():
     # is dropped!
 
 
+def test_dict_key_error():
+    try:
+        Schema({'k': int}).validate({'k': 'x'})
+    except SchemaError as e:
+        assert e.code == "Key 'k' error:\n'x' should be instance of 'int'"
+    try:
+        Schema({'k': {'k2': int}}).validate({'k': {'k2': 'x'}})
+    except SchemaError as e:
+        code = "Key 'k' error:\nKey 'k2' error:\n'x' should be instance of 'int'"
+        assert e.code == code
+    try:
+        Schema({'k': {'k2': int}}, error='k2 should be int').validate({'k': {'k2': 'x'}})
+    except SchemaError as e:
+        assert e.code == 'k2 should be int'
+
+
 def test_complex():
     s = Schema({'<file>': And([Use(open)], lambda l: len(l)),
                 '<path>': os.path.exists,
@@ -511,3 +576,19 @@ def test_copy():
     s2 = copy.deepcopy(s1)
     assert s1 is not s2
     assert type(s1) is type(s2)
+
+
+def test_inheritance():
+    def convert(data):
+        if isinstance(data, int):
+            return data + 1
+        return data
+
+    class MySchema(Schema):
+        def validate(self, data):
+            return super(MySchema, self).validate(convert(data))
+
+    s = {'k': int, 'd': {'k': int, 'l': [{'l': [int]}]}}
+    v = {'k': 1, 'd': {'k': 2, 'l': [{'l': [3, 4, 5]}]}}
+    d = MySchema(s).validate(v)
+    assert d['k'] == 2 and d['d']['k'] == 3 and d['d']['l'][0]['l'] == [4, 5, 6]
